#!/usr/bin/env python
import numpy as np
import matplotlib
matplotlib.use('Agg')
import matplotlib.pyplot as plt
from matplotlib import gridspec
from matplotlib.ticker import MaxNLocator


# plotting tool
# skip to main()

#auxiliary function for text formatting
def initializetex():
	#labels style
	#font.serif         : Times, Palatino, New Century Schoolbook, Bookman, Computer Modern Roman
	#font.sans-serif    : Helvetica, Avant Garde, Computer Modern Sans serif
	#font.cursive       : Zapf Chancery
	#font.monospace     : Courier, Computer Modern Typewriter
	plt.rc('font',**{'family':'serif','serif':['Computer Modern Roman']})
	plt.rc('font',**{'family':'sans-serif','sans-serif':['Computer Modern Sans serif']})
	plt.rc('text', usetex=True)
	plt.rc('font', family='serif')
	plt.rcParams.update({'font.size':12})
	plt.rcParams['axes.labelsize'] = plt.rcParams['font.size']
	plt.rcParams['axes.titlesize'] = 1.2*plt.rcParams['font.size']
	plt.rcParams['legend.fontsize'] = plt.rcParams['font.size']
	plt.rcParams['xtick.labelsize'] = plt.rcParams['font.size']
	plt.rcParams['ytick.labelsize'] = plt.rcParams['font.size']
	params = {'legend.fontsize': 12,
	'legend.handlelength': 2}
	plt.rcParams['hatch.linewidth'] = 0.5  # previous svg hatch linewidth
	plt.rcParams.update(params)

#figure size optimization
def figsize(fig_dims):

	golden_ratio  = (np.sqrt(5) - 1.0) / 2.0  # because it looks good

	fig_width_in  = 3.5  # figure width in inches
	fig_height_in = fig_width_in * golden_ratio   # figure height in inches
	fig_dims    = [fig_width_in, fig_height_in] # fig dims as a list
	return fig_dims

#plot style
def style(ax1):
	almost_black = '#262626'
	light_grey = np.array([float(248)/float(255)]*3)
	# For remaining spines, thin out their line and change the black to a slightly off-black dark grey
	spines_to_keep = ['bottom', 'left','top','right']
	for spine in spines_to_keep:
		ax1.spines[spine].set_linewidth(0.5)
		ax1.spines[spine].set_color(almost_black)

	# Change the labels to the off-black
	ax1.xaxis.label.set_color(almost_black)
	ax1.yaxis.label.set_color(almost_black)
	ax1.tick_params(which='both',colors=almost_black)

	# Change the axis title to off-black
	ax1.title.set_color(almost_black)
	grid(ax1)


def grid(ax1):
	grid_grey = '#C7C7C7'
	#grid style
	ax1.set_axisbelow(True)
	ax1.grid(color=grid_grey, linestyle='-', linewidth=0.25)

def despine(ax):
	spines_to_remove = ['top', 'right']
	for spine in spines_to_remove:
		ax.spines[spine].set_visible(False)
	almost_black = '#262626'
	spines_to_keep = ['bottom', 'left']
	for spine in spines_to_keep:
		ax.spines[spine].set_linewidth(0.5)
		ax.spines[spine].set_color(almost_black)
	ax.xaxis.set_ticks_position('bottom')
	ax.yaxis.set_ticks_position('left')

def despineall(ax):
	spines_to_remove = ['top', 'right', 'bottom', 'left']
	for spine in spines_to_remove:
		ax.spines[spine].set_visible(False)

#legend style
def legstyle(leg):
	almost_black = '#262626'
	light_grey = np.array([float(248)/float(255)]*3)
	# Change the legend label colors to almost black, too
	texts = leg.get_texts()
	leg.get_frame().set_edgecolor(almost_black)
	for t in texts:
		t.set_color(almost_black)
	rect = leg.get_frame()
	rect.set_facecolor(light_grey)
	rect.set_linewidth(0.0)