      function outside_cuts(p)
      implicit none

      include 'types.f'
      include 'mxpart.f'
      include 'npart.f'
      include 'constants.f'
      include 'limits.f'
      include 'energy.f'
      include 'cuts.f'
      include 'mcfmparameters.f'
      include 'nproc.f'
      include 'nprocesses.f'
      include 'hdecaymode.f'
      
      real(dp):: x1,x2
      real(dp):: p(mxpart,4)
      logical:: outside_cuts

      real(dp):: pl1(4), pl2(4), pg1(4), pg2(4), pgg(4), pll(4)
      real(dp):: y1,y2,eta1,eta2,pt1,pt2,m1,m2
      real(dp):: ygg,etagg,ptgg,mgg
      real(dp):: yll,etall,ptll,mll
      real(dp) :: mT, ETmiss
      real(dp):: pnu(4),pel(4)
c---------------> start
      outside_cuts=.false.

      x1=-2._dp*p(1,4)/sqrts
      x2=-2._dp*p(2,4)/sqrts

!     invariant mass window
      if( x1*x2*sqrts**2 .lt. wsqmin ) outside_cuts = .true.
      if( x1*x2*sqrts**2 .gt. wsqmax ) outside_cuts = .true.
c-----------------------------------------------------------------------
      if (use_cuts) then
         
!     cuts for Z+jets as in ATLAS measurements
         if (nproc == nDY) then
            pl1(1:4) = p(3,1:4)
            pl2(1:4) = p(4,1:4)
            pll(1:4) = p(3,1:4) + p(4,1:4)
            
            call getyetaptmass(pl1, y1,eta1,pt1,m1)
            call getyetaptmass(pl2, y2,eta2,pt2,m2)
            call getyetaptmass(pll, yll,etall,ptll,mll)

!            transverse momentum cut
            !if( pt1 .lt. 25d0) outside_cuts = .true.
            !if( pt2 .lt. 25d0) outside_cuts = .true.
            ! cuts from ATLAS 13 TeV analysis 2019
            if( pt1 .lt. 27d0) outside_cuts = .true.
            if( pt2 .lt. 27d0) outside_cuts = .true.

!            rapidity cut on the leptons
            if (abs(eta1) .ge. 2.5d0) outside_cuts = .true.
            if (abs(eta2) .ge. 2.5d0) outside_cuts = .true.

!           rapidity cut on the virtual boson
            !if (abs(yll) .lt. born_eta_min .or. abs(yll) .gt. born_eta_max) outside_cuts = .true.
         end  if

!     cuts for W processes
         if (nproc == nWp) then
c           this has been benchmarked
c           try with different seeds and high stat as it seems that sometimes 
c           the vegas error is not accurate

c           For nwz=+1
c           u(-p1)+dbar(-p2)-->W^+(n(p3)+e^+(p4))
            pnu(1:4) = p(3,1:4)
            pel(1:4) = p(4,1:4)
            pll(1:4) = p(3,1:4) + p(4,1:4)

            call getyetaptmass(pel,y1,eta1,pt1,m1)
            call getyetaptmass(pnu,y2,eta2,pt2,m2)

            !charged lepton cuts (pt and rapidity)
            if( pt1 .lt. 25d0 ) outside_cuts = .true.
            if( abs(eta1) .ge. 2.5d0 ) outside_cuts = .true.

            !transverse mass cut
            mT = sqrt((abs(pt1)+abs(pt2))**2 - sqrt(pll(1)**2 + pll(2)**2))
            if(mT .lt. 50d0 ) outside_cuts = .true.

            !missing transverse energy
            ETmiss = sqrt(m2**2+pt2**2)
            if(ETmiss .lt. 25d0 ) outside_cuts = .true.

         elseif (nproc == nWm) then
c           this has been benchmarked
c           try with different seeds and high stat as it seems that sometimes 
c           the vegas error is not accurate

C           For nwz=-1
c           d(-p1)+ubar(-p2)-->W^-(e^-(p3)+nbar(p4))
            pnu(1:4) = p(4,1:4)
            pel(1:4) = p(3,1:4)
            pll(1:4) = p(3,1:4) + p(4,1:4)

            call getyetaptmass(pel,y1,eta1,pt1,m1)
            call getyetaptmass(pnu,y2,eta2,pt2,m2)

            !charged lepton cuts (pt and rapidity)
            if( pt1 .lt. 25d0 ) outside_cuts = .true.
            if( abs(eta1) .ge. 2.5d0 ) outside_cuts = .true.

            !transverse mass cut
            mT = sqrt((abs(pt1)+abs(pt2))**2 - sqrt(pll(1)**2 + pll(2)**2))
            if(mT .lt. 50d0 ) outside_cuts = .true.

            !missing transverse energy
            ETmiss = sqrt(m2**2+pt2**2)
            if(ETmiss .lt. 25d0 ) outside_cuts = .true.
         end if

!     cuts for H (-> 2y) +jets as in ATLAS measurement
         if (nproc == nH .and. hdecaymode == 'gaga') then
            pg1(1:4) = p(3,1:4)
            pg2(1:4) = p(4,1:4)
            pgg(1:4) = p(3,1:4) + p(4,1:4)
            call getyetaptmass(pg1, y1,eta1,pt1,m1)
            call getyetaptmass(pg2, y2,eta2,pt2,m2)
            call getyetaptmass(pgg, ygg,etagg,ptgg,mgg)

            if (min(pt1,pt2).le.31.25d0) outside_cuts = .true.
            if (max(pt1,pt2).le.43.75d0) outside_cuts = .true.

            if (abs(eta1).ge.1.37d0 .and. abs(eta1).le.1.52d0) outside_cuts = .true.
            if (abs(eta1).ge.2.37d0) outside_cuts = .true.
            
            if (abs(eta2).ge.1.37d0 .and. abs(eta2).le.1.52d0) outside_cuts = .true. 
            if (abs(eta2).ge.2.37d0) outside_cuts = .true.
            
            if (abs(ygg).ge.2.37d0) outside_cuts = .true.

            ! keep an on-shell Higgs to match NNLOJET
            if (mgg.lt.124.9 .or. mgg.gt.125.1) outside_cuts = .true.
         end if
         
      end if
c-----------------------------------------------------------------------
        
      


      return
      end
      
c#######################################################################
      subroutine getyetaptmass(p,y,eta,pt,mass)
      implicit none
      include 'types.f'
      include 'constants.f'
      
      real(dp) p(4),y,eta,pt,mass,pv
      real(dp) tiny
      parameter (tiny=1.e-5_dp)
      y=0.5_dp*log((p(4)+p(3))/(p(4)-p(3)))
      pt=sqrt(p(1)**2+p(2)**2)
      pv=sqrt(pt**2+p(3)**2)
      if(pt<tiny)then
         eta=sign(one,p(3))*1.e8_dp
      else
         eta=0.5_dp*log((pv+p(3))/(pv-p(3)))
      endif
      mass=sqrt(abs(p(4)**2-pv**2))
      end

