module dynamical_scale
  use types; use consts_dp
  use emsn_tools
  use rad_tools
  use warnings_and_errors

  
  public:: set_dynamical_scale, singlet_invariant_mass

contains

  ! this function calculates dynamical scales in the Born kinematics
  function set_dynamical_scale(pborn) result(res)
    include 'mxpart.f'
    real(dp), intent(in) :: pborn(mxpart,4)
    real(dp) :: res, pZ(4), pW(4)

    if( trim(cs%proc)=='DY' )then
       pZ(1:4) = pborn(3,1:4) + pborn(4,1:4)
       !--- invariant mass of Z-boson
       res=sqrt( abs(dot(pZ,pZ)) )
    elseif ( trim(cs%proc)=='Wp' .or. trim(cs%proc)=='Wm'  )then
       pW(1:4) = pborn(3,1:4) + pborn(4,1:4)
       !--- invariant mass of Z-boson
       res=sqrt( abs(dot(pW,pW)) )
    else
       call wae_error('set_dynamical_scale: no dynamical scale implemented for this process')
    endif
    
  end function set_dynamical_scale


  ! this function calculates dynamical scales in the Born kinematics
  function singlet_invariant_mass(pborn) result(res)
    include 'mxpart.f'
    real(dp), intent(in) :: pborn(mxpart,4)
    real(dp) :: res, psinglet(4)

    psinglet(1:4) = pborn(1,1:4) + pborn(2,1:4)
    !--- invariant mass of color singlet:
    res=sqrt(abs( dot(psinglet,psinglet) ))

  end function singlet_invariant_mass

  
  

end module dynamical_scale
