! module containing global defition of SC emissions and base vectors including recoil in different cases 
! contains also routines to generate emission, recoil, and addition/subtraction os SC momenta 

module momenta_and_recoil
  use types; use consts_dp
  use random_dists
  use warnings_and_errors
  use rad_tools
  implicit none

  private 


  ! scmom is a soft-collinear momentum, specified by rapidity and kt 
  type scmom
     real(dp) :: ktmod, rap, kt(2)
  end type scmom
  public :: scmom

  ! addition and subtraction of scmom used for recombination and recoil 
  ! addition of kt(2) vectorially, assigns rapidity of emission with largest abs(rap) 
  interface operator(+)
     module procedure scmom_add
  end interface

  ! subtracts of kt(2) vectorially, assigns rapidity of emission with largest abs(rap) 
  interface operator(-) 
     module procedure scmom_sub
  end interface

  public :: operator(+), operator(-)


  integer, parameter, public :: maxems = 10000000
  ! q contais soft+collinear emissions + special emission 
  ! qbase_no1st contain base vectors with recoils from 2:nems SC emissions 
  ! qbase contain base vectors with recoils from 1:nems SC emissions 
  ! qbase_sp contain base vectors with recoils from 1:nems SC emissions and 
  ! from special emission 
  type(scmom), public :: q(maxems+2) ! it must allocate two special emissions
  integer,     public :: nems
  public :: init_scmom, gen_scmom

contains

  !---------------------------------------------------------------------
  ! assumption that differences between rapidities are much
  ! larger than any change in angle that would be caused by change in kt
  ! 
  ! So in sum just keep larger absolute rapidity
  function scmom_add(p1,p2) result(p3)
    type(scmom), intent(in) :: p1,p2
    type(scmom)             :: p3
    if (abs(p1%rap) > abs(p2%rap)) then
       p3%rap = p1%rap
    else
       p3%rap = p2%rap
    end if
    p3%kt = p1%kt + p2%kt
    p3%ktmod = sqrt(p3%kt(1)**2 + p3%kt(2)**2)
  end function scmom_add

  ! and in subtraction just keep larger absolute rapidity
  function scmom_sub(p1,p2) result(p3)
    type(scmom), intent(in) :: p1,p2
    type(scmom)             :: p3
    if (abs(p1%rap) > abs(p2%rap)) then
       p3%rap = p1%rap
    else
       p3%rap = p2%rap
    end if
    p3%kt = p1%kt - p2%kt
    p3%ktmod = sqrt(p3%kt(1)**2 + p3%kt(2)**2)
  end function scmom_sub



 
  ! given each lnv and ktpow, generate soft emission (scmom) 
  subroutine gen_scmom(v,ktpow,pmom) 
    real(dp), intent(in) :: v
    real(dp), intent(in) :: ktpow
    type(scmom), intent(inout) :: pmom

    pmom%rap = two*ran()-one
    !pmom%ktmod = v**(one/ktpow)
    !pmom%kt = rphi_2d(pmom%ktmod)
    ! generate first the azimuthal direction, useful for the phi*
    pmom%kt=rphi_2d(one) ! returns a 2d vector of length 1 with a random direction
    pmom%ktmod = v**(one/ktpow)
    pmom%kt=pmom%kt*pmom%ktmod

  end subroutine gen_scmom

  ! initializze a soft-collinear momentum to kt=0 and fixed, 
  ! initial rapidity 
  elemental subroutine init_scmom(p,init_rap)
    type(scmom), intent(inout) :: p
    real(dp), intent(in)     :: init_rap

    p%rap = init_rap
    p%ktmod = zero
    p%kt = zero

  end subroutine init_scmom
end module momenta_and_recoil
